/** \mainpage 
    v1.1 MIDI sample code
    
    Implemetation of a single USB Midi device on a a CH552. It uses uart1 for
    Midi In and Midi Out on two 128 byte fifo ringbuffers. It will work on 
    any controller CH551 - CH554.
     - Rx data will be parsed and send over the EP In to the host. 
     - Tx data will be read from EP out and put into the Tx fifo. 
    
    Both fifos will be handled in IRQ mode. Running status will be maintained
    for input and output.

    Since CH552 has 2 uarts its possible to build a device with 2 Midi Ports.
    Generally there are 2 ways to this:
      - use a second EP pair
      - use the same EP pair for both ports with Cable Index No (CIN)
   
    For windows systems its best to use a second bulk pair. 
    Unfortunately the last time I checked it, MacOs does only support shared
    EPs for multible Midi Ports. If you use shared Eps on Win there might be
    a chance to loose some Sysex transmissions.
    
    The device will be recognized starting from Win XP or any linux system as
    well as MacOS as a Midi Device with one MidiIn and one MidiOut. 
    Don't connect it to a W2000 system. W2000 will crash, W98 will hang and 
    Win Me has just very limited support for usbmidi.
    
    Midi IO is by default done on the second uart and there with the alternte 
    pin mapping. Two leds will indicate if there is trafic on the ports. 
    The Midi out will send Active Sensing if there is no other activity for 320 ms.

    Currently there are three configuration options available within the DataFlash:
     - DataFlash{0]: 0 Active Sensing is off
     - DataFlash[1]: 0 Running Status for Output is off
     - DataFlash[2]: 0 Midi traffic on the first Uart

    Any other value will activate the option, so by default all options are 
    active. For CH551 the third option will be ignored and Uart0 will be used.
  
    The code is ready to be compiled in Keil or SDCC. See the prj_keil and 
    prj_sdcc directory for details.
    
    Tools: 
    midiox old but still the best utility for midi 
    see http://www.midiox.com

    Todo:
    - create correct USB_GET_STATUS requests now it simply returns 0x0000
    - check the firmware against UsbCv for Spec conformance 
    - use SysEx to change the Config Options
    - implement FW update using SysEx

*/

#include ".\..\inc\ch554.h"
#include ".\..\inc\mytypes.h"
#include ".\..\inc\usb\usb11.h"
#include ".\..\inc\device.h"
#include ".\..\inc\uart.h"
#include ".\..\inc\descriptor.h"
//#include <string.h>                     // memcpy

#ifdef __C51__
  xdata uint8_t EP0_Buffer   [EP0_SIZE]     _at_ 0x0000; // control
  xdata uint8_t EP2_OutBuffer[BULK_SIZE]    _at_ 0x0040; // bulk out
  xdata uint8_t EP2_InBuffer [BULK_SIZE]    _at_ 0x0080; // bulk in
  xdata uint8_t MidiRxBuffer [MIDIBUF_SIZE] _at_ 0x0100; // fifo rx
  xdata uint8_t MidiTxBuffer [MIDIBUF_SIZE] _at_ 0x0200; // fifo tx
#elif __SDCC_mcs51
  __xdata __at (0x000) uint8_t EP0_Buffer   [EP0_SIZE];
  __xdata __at (0x040) uint8_t EP2_OutBuffer[BULK_SIZE];
  __xdata __at (0x080) uint8_t EP2_InBuffer [BULK_SIZE];
  __xdata __at (0x100) uint8_t MidiRxBuffer [MIDIBUF_SIZE];
  __xdata __at (0x200) uint8_t MidiTxBuffer [MIDIBUF_SIZE];
#else
  #error unknown compiler 
#endif

// usb vars
const uint8_t *pDescr;   /**< pointer to a descriptor part */
tSETUP    Setup;         // holding the setup packet
uint8_t   UsbConfig;     /**< config value here 1 after the device is configured */
uint8_t   UsbState;      /**< the actual enum state */
bit       bRequestError; /**< indicates any unknown or unsupported request */

//  midi0 vars 
tMSG     msg0;           /**< struct holding incomming midi messages */


volatile tFIFOCTL MidiRxCtl;      // fifos for rx and tx
volatile tFIFOCTL MidiTxCtl;


uint8_t  inStatus0;      /**< incomming running status */
uint8_t  outStatus0;     /**< outgoing running status experimental */
uint8_t  usbIdx0;        /**< epin buffer index */ 
uint8_t  CfgOutStat;     /**< Option for switching on out Running Status */ 
uint8_t  CfgSense;       /**< Option for switching on Active Sensing*/
uint8_t  CfgUart1;       /**< Option for selecting Uart */

bit      bRaw0MidiMode;  /**< true when the device is in raw midi mode */
bit      bMsg0Complete;  /**< true when msg contains a complete midi packet */
bit      bUsb0InReady;   /**< true when the bulk in can aaccept data again*/
bit      bEP2Lock;       /**< true after a transmisson on EP2Out */

/** Initialize the usb core

    Enable the EP0 (the default ep. Initialize the DMA pointers but leave 
    the additional EPs disabled since before SetConfigure no EPs besides 
    EP0 are allowed. Init the usb irqs for reset, suspend and transfer 
    and set irq  priority to low allowing both uart irqs to interrupt 
    a usb irq. Finally enable the core (connect).
    \note: EA remains unchanged!
    
*/
void InitUsbDevice(void)
{
   USB_CTRL    = 0;
   UEP2_3_MOD  = bUEP2_RX_EN | bUEP2_TX_EN;
   UEP0_DMA_H  = ((uint16_t) &EP0_Buffer) >> 8;
   UEP0_DMA_L  = ((uint16_t) &EP0_Buffer) & 0xFF;
   UEP2_DMA_H  = ((uint16_t) &EP2_OutBuffer) >> 8;
   UEP2_DMA_L  = ((uint16_t) &EP2_OutBuffer) & 0xFF;
   USB_CTRL    = bUC_DEV_PU_EN | bUC_INT_BUSY | bUC_DMA_EN;
   UDEV_CTRL   = bUD_PD_DIS | bUD_PORT_EN;
   USB_INT_FG |= 0x1F; 
   USB_INT_EN |= bUIE_SUSPEND | bUIE_TRANSFER | bUIE_BUS_RST;
   IP_EX      &= ~bIP_USB;  //low priority for USB 

   IE_USB = 1; 
   UsbState = USB_STATE_POWERED;     
}

/** Initialize all components

    - Init the uart to 31250 baud for midi
    - Init T2 for timeouts  
    - initialize the usb core 
*/
void InitDevice(void)
{
   EA = 0;
   LED1=LED_OFF;
   LED2=LED_OFF; 
   PIN_FUNC |= bUART1_PIN_X;
   if(CfgUart1) InitUart1();
   else         InitUart0();
   IntitTimeout();
   InitUsbDevice();

}

//***************************************************************************
//***  midi code
//***************************************************************************

// decoder tables
code uint8_t MsgCtlCmd[0x07] = {                   3,3,3,3, 2,2,3   }; // 8x..Ex
code uint8_t MsgSysCmd[0x10] = { 3,2,3,2, 0,0,1,0, 1,0,1,1, 1,0,1,1 }; // F0..FF
code uint8_t CidLen   [0x10] = { 0,0,2,3, 3,1,2,3, 3,3,3,3, 2,2,3,1 }; // Table 4-1 usb Midi Spec
code uint8_t CinTbl   [0x03] = { 5,2,3 };

/** Get the length of a midi command 

    Different commands have different sets of parameter. This function
    reads the command and determins how many bytes will follow.
    \param cmd  midistatus byte (0x80..0xFF)
    \result     number of data bytes for that cmd (0..3) 0 = unsupported
*/
uint8_t MidiMsgLen(uint8_t cmd)
{
   uint8_t msgid = cmd >> 4 ;       // just check the upper nibble

   if (msgid == 0x0F) return MsgSysCmd[cmd & 0x0F]; // 0xF0 ..0xFF
   return MsgCtlCmd[msgid-8];                       // 80..EF
}


/** Get the code index number
   
    The code index number is the low nibble of the first byte within the
    midi packet. See table 4.1 of the midi 1.0 spec. The function handles 
    running status as well as these special sysex codes. 
    \return  CIN (0x00 .. 0x0F)
*/
uint8_t  Cin(void)
{
   if (inStatus0  < 0xf0) return (inStatus0>>4);                                      // 8x .. Ex 
   if (inStatus0 == 0xf0) return 4 + ((msg0.Buf[msg0.Cnt-1] == 0xf7) ? msg0.Cnt : 0); // F0
   if (inStatus0  < 0xf8) return CinTbl[msg0.Cnt-1];                                  // F1 .. F7
   return 0x0F;             // single byte                                            // F8 .. FF  
}


/** Pass a midi command to the EPIn buffer

    Put a packet into the buffer without any changes. Used for real time 
    messages or for raw midimode. There is no parsing done on that bytes
    just cable id and CIN = 0x0F is added in the first byte
    \param[in] rx the byte to put into byte1
    \param[in] cid cable index to be used here usually 0
*/     
void PassPacket0(uint8_t rx, uint8_t cid)
{
   EP2_InBuffer[usbIdx0++] = 0x0F | (cid<<4); // singleByte Message on cid
   EP2_InBuffer[usbIdx0++] = rx;              //  
   EP2_InBuffer[usbIdx0++] = 0;               // others =0;
   EP2_InBuffer[usbIdx0++] = 0;
   bMsg0Complete=1;
}


/** Put a message into the bulk in buffer

    Store the four bytes of a usb midi cmd into the EP In buffer and
    reset the msg to 0. Maintain a running status since incomming may or
    may not use running status.
    \param[in] cid cable index number here usually 0
*/    
void PutInMessage0(uint8_t cid)
{
   if (MidiMsgLen(inStatus0))         // ignore the packet if length = 0
   {
      EP2_InBuffer[usbIdx0++] = Cin() | (cid<<4); // CableIndex und CID
      EP2_InBuffer[usbIdx0++] = msg0.Buf[0];      // 1. Byte
      EP2_InBuffer[usbIdx0++] = msg0.Buf[1];      // 2. Byte
      EP2_InBuffer[usbIdx0++] = msg0.Buf[2];      // 3. Byte 
      bMsg0Complete=1;
      msg0.Cnt = msg0.Buf[0] = msg0.Buf[1] = msg0.Buf[2] = 0; // clear the struct
      if (inStatus0 < 0xf0) msg0.Buf[msg0.Cnt++] = inStatus0; // remember running status
      if (inStatus0 > 0xf0) inStatus0 = 0;                    // clear any running status on  sys cmds
   }
}

/** Process incoming bytes from the uart stream

    Form usb midi messages from the incomming bytes. if in raw mode or if its 
    a single byte cmd simply create a new message. If there is currently a 
    sysex running terminate the sysex (cancel) first and start building a 
    usb midi message. 
    A global flag will indicate if msg contains a complete packet.
    \param[in] rx  the byte from the input stream
    \param[in] cid cable index to add
*/
void ProcessByte0(uint8_t rx, uint8_t cid)
{
   bMsg0Complete=0;
   //
   // test if in raw mode
   //
   if (bRaw0MidiMode) { PassPacket0(rx,cid); return; }
   //
   // pass supportet realtime message
   //
   if (rx > 0xF7)                   
   { 
      if (MidiMsgLen(rx)) PassPacket0(rx,cid);  
      bMsg0Complete=1;  // ignore unsupported
      return;                       
   }
   // test for new status bytes
   if (rx & 0x80)                     
   {
   	  // if there is a sysex runing terminate it first
      if (inStatus0 ==0xF0) {msg0.Buf[msg0.Cnt++] = 0xF7; PutInMessage0(cid);}    
      // update running status
      inStatus0 = rx;
      msg0.Cnt = 0;
   } 
   
   // collect the the additional bytes
   if (inStatus0) 
   {
      msg0.Buf[msg0.Cnt++] = rx;          
      if (msg0.Cnt >= MidiMsgLen(inStatus0)) PutInMessage0(cid);           
   }
}

/** Transfer incomming data to the EP IN buffer
    
    Read data from the rx fifo and put them into RX EP. Terminate the transer 
    if one of the following contitions occurs:
    - EP is no longer ready to accept data
    - there are no more data in the fifo
    - the EP is full
    
    If the last message is not complete exit. The transfer will be restarted
    later again. Unused parts of the EP buffer will be padded with 0
    and the Ep will be armed with 0x40 bytes.
*/
void TransferMidi2Usb0(void)
{
   uint8_t rx;

   while ((bUsb0InReady)     &&    // bufferready
          (MidiRxCtl.level)  &&    // still bytes in the stream 
          (usbIdx0 < BULK_SIZE))   // usb not full
   { 
      EA = 0;
      rx = MidiRxBuffer[MidiRxCtl.wr++]; // pick a byte
      MidiRxCtl.wr &= MIDIBUF_MASK;
      MidiRxCtl.level--;
      EA = 1;
      ProcessByte0(rx,0x00);              // and process it
   }
   // usb needs to have complete packets
   if (!bMsg0Complete) return;         
   
   if ((usbIdx0) && (bUsb0InReady)) 
   {
      bUsb0InReady = 0;                 // lock buffer
      // fill the rest with zero 
      while (usbIdx0 < BULK_SIZE)            
      {
         EP2_InBuffer[usbIdx0++]=0;
      }
      usbIdx0=0;
      //Arm(Ep2,BULK_SIZE)
      UEP2_T_LEN  = BULK_SIZE;
      UEP2_CTRL  &= ~MASK_UEP_T_RES;
   }
}
#ifdef __SDCC_mcs51
/*
 bugbug SDCC can not put interrupts into the vector table
 which are declared elsewhere. 
 Quote from the doc:
  "If you have multiple source files in your project, interrupt 
   service routines can be present in any of them, but a prototype 
   of the isr MUST be present or included in the file that contains 
   the function main."
*/
extern void TimeoutIrq (void) interrupt INT_NO_TMR2;
extern void Uart1_ISR(void) interrupt INT_NO_UART1;
extern void Uart0_ISR(void) interrupt INT_NO_UART0;
#endif

void Usb_IRQ (void) interrupt INT_NO_USB using 1
{  
   uint8_t rxlen; 
   uint8_t i;
   uint8_t status0=0;  // init to zero because of SDCC warning
   uint8_t len=0;      // todo this just works for descriptor sets < 256 bytes 
   if(UIF_TRANSFER)
   {  // transfer interrupt
      switch (USB_INT_ST & (MASK_UIS_TOKEN | MASK_UIS_ENDP))
      {
         
         case UIS_TOKEN_IN | 2:
              {  //usrt1 rx
                 UEP2_T_LEN   = 0;
                 UEP2_CTRL    = UEP2_CTRL & ~ MASK_UEP_T_RES | UEP_T_RES_NAK;
                 bUsb0InReady = 1;
              }
              break;
         case UIS_TOKEN_OUT | 2:
              {  // Midi tx
                 if ( U_TOG_OK )
                 {
                 	  UEP2_CTRL = UEP2_CTRL & ~ MASK_UEP_R_RES | UEP_R_RES_NAK;
                    rxlen = USB_RX_LEN; // should be always 0x40
                    i=0;
                    while (i < rxlen)
                    {
                    	 len = CidLen[EP2_OutBuffer[i++] & 0x0F];
                       if(CfgOutStat) //outRunning status is configurable
                       { 
                          if(len) 
                          {
                             status0 = EP2_OutBuffer [i];
                             if (status0 >= 0xF0) status0 = 0;
                             if(outStatus0==status0)
                             {
                                len--;
                                i++;  
                             }   
                          }
                       }    
                  	   while (len)
                  	   { 
                  	 	    MidiTxBuffer[MidiTxCtl.rd++] = EP2_OutBuffer[i++];
                  	 	    MidiTxCtl.rd&=MIDIBUF_MASK;
                  	 	    MidiTxCtl.level++;	     
                  	 	    len--; 
                  	   }
                    	 if (i & 0x03) { i &= ~0x03; i+=4;} // adjust index
                       outStatus0=status0;
                    }

                    if ((MidiTxCtl.level) && (bTx0Empty))
                    {
                    	 bTx0Empty=0;
                    	 U1TI=1;
                    }
                    bEP2Lock=1;	                    
                 } 
              }
              break;
         case UIS_TOKEN_SETUP | 0:
              {               
                 UEP0_CTRL &=0xF2; // remove any possible STALL contition from previous request 
                 if(USB_RX_LEN == (sizeof(tSETUP)))
                 {
                    // copy the setup packet
                    // USB is lsb first, keil is msb first
                    Setup.bRequestType = EP0_Buffer[0];
                    Setup.bRequest     = EP0_Buffer[1];
                    Setup.wValueLo     = EP0_Buffer[2]; 
                    Setup.wValueHi     = EP0_Buffer[3];
                    Setup.wIndexLo     = EP0_Buffer[4]; 
                    Setup.wIndexHi     = EP0_Buffer[5];

                    Setup.wLength       = EP0_Buffer[7]; 
                    Setup.wLength <<=8;
                    Setup.wLength      |= EP0_Buffer[6];

                    bRequestError=0;
                    len = 0;
                    //
                    // chck the std recipient
                    //
                    if ((Setup.bRequestType & _REQUESTMASK) == _STANDART)
                    {  
                       switch(Setup.bRequest)
                       {
                          case USB_GET_DESCRIPTOR:
                               switch(Setup.wValueHi) 
                               {
                                  case USB_DEVICE_DESCRIPTOR: 
                                       {                                     
                                          pDescr = &DevDesc[0];
                                          len    = DevDesc[0];
                                          //len = sizeof(DevDesc);
                                       }
                                       break;
                                  case USB_CONFIGURATION_DESCRIPTOR:
                                       {
                                           pDescr = &CfgDesc[0];
                                           len    = CfgDesc[2];
                                           //len = sizeof(CfgDesc);
                                       }  
                                       break;
                                  case USB_STRING_DESCRIPTOR:
                                       {
                                          switch (Setup.wValueLo)
                                          {
                                             case 0:
                                                  pDescr = &LangDes[0];
                                                  len    = LangDes[0];  
                                                  //len = sizeof(LangDes);
                                                  break;
                                             case 1:
                                                  pDescr = &Manuf_Des[0];
                                                  len    = Manuf_Des[0];
                                                  //len = sizeof(Manuf_Des);
                                                  break;
                                             case 2:
                                                  pDescr = &Prod_Des[0];
                                                  len    = Prod_Des[0];
                                                  //len = sizeof(Prod_Des);
                                                  break;
                                             default:
                                                  bRequestError=1;
                                                  //break;              
                                          }
                                       }
                                       break;
                                  default:
                                       bRequestError=1;
                                       //break;
                               }
                               if (len < Setup.wLength) Setup.wLength = len;
                               len = (len >= EP0_SIZE) ? EP0_SIZE 
                                                       : len;
                               //memcpy(EP0_Buffer,pDescr,len);
                               //pDescr += len;  
                               for (i=0; i<len; i++) {EP0_Buffer[i] = *pDescr++;}
                               Setup.wLength -= len;
                               break;
                          case USB_SET_ADDRESS:
                               break;
                          case USB_GET_CONFIGURATION:
                               EP0_Buffer[0] = UsbConfig;
                               len = 1;
                               break;
                          case USB_SET_CONFIGURATION:
                               {
                                  UsbConfig = Setup.wValueLo;
                                  if (UsbConfig)
                                  { 
                                     //UEP2_3_MOD = bUEP2_RX_EN | bUEP2_TX_EN;
                                     UEP2_CTRL  = bUEP_AUTO_TOG | UEP_T_RES_NAK | UEP_R_RES_ACK;
                                     bTx0Empty  = 1;
                                     bUsb0InReady = 1; 
                                     UsbState   = USB_STATE_CONFIG;
                                     MidiRxCtl.rd = MidiRxCtl.wr = MidiRxCtl.level=0;
                                     break;
                                  }   
                                  bTx0Empty  = 0;
                                  //UEP2_3_MOD = 0;
                                  UsbState   = USB_STATE_ADDRESSED; 
                               }
                               break;
                          case USB_GET_INTERFACE:
                               break;
                          case USB_CLEAR_FEATURE: 
                               if(( Setup.bRequestType & _RECIPIENTMASK) == _DEVICE ) 
                               {
                                   if(Setup.wValueLo == FEATURE_REMOTE_WAKEUP )
                                   {
                                       if( CfgDesc[7] & 0x20 )
                                       {

                                       }
                                       else bRequestError=1; 
                                   }
                                   else bRequestError=1;
                               }
                               else if ((Setup.bRequestType & _RECIPIENTMASK) == _ENDPOINT )
                               {
                                  switch (Setup.wIndexLo &0x0F)
                                  {  
                                     case 0x02:
                                          if (Setup.wIndexLo & _HOST) 
                                             UEP2_CTRL = UEP2_CTRL & ~ ( bUEP_T_TOG | MASK_UEP_T_RES ) | UEP_T_RES_NAK;
                                          else
                                             UEP2_CTRL = UEP2_CTRL & ~ ( bUEP_R_TOG | MASK_UEP_R_RES ) | UEP_R_RES_ACK;
                                          break;
                                     default:
                                          bRequestError=1; 
                                         //break;
                                  }
                               }
                               else bRequestError=1; 
                               break;
                          case USB_SET_FEATURE: 
                               if((Setup.bRequestType & 0x1F) == _DEVICE) 
                               {
                                   if(Setup.wValueLo == FEATURE_REMOTE_WAKEUP )
                                   {
                                       if( CfgDesc[7] & 0x20 )
                                       {
                                           while ( XBUS_AUX & bUART0_TX )
                                           {
                                              ;
                                           }
                                           SAFE_MOD = 0x55;
                                           SAFE_MOD = 0xAA;
                                           WAKE_CTRL = bWAK_BY_USB | bWAK_RXD0_LO | bWAK_RXD1_LO;                      //USBRXD0/1źʱɱ
                                           PCON |= PD;                                                                 //˯
                                           SAFE_MOD = 0x55;
                                           SAFE_MOD = 0xAA;
                                           WAKE_CTRL = 0x00;
                                       }
                                       else bRequestError=1;
                                   }
                                   else bRequestError=1;
                               }
                               else if((Setup.bRequestType & 0x1F) == _ENDPOINT ) 
                               {
                                   if(Setup.wValueLo == FEATURE_ENDPOINT_HALT )
                                   {
                                      switch(Setup.wIndexLo & 0x0F)
                                      {  
                                         case 0x02:
                                              if (Setup.wIndexLo & _HOST) 
                                                 UEP2_CTRL = UEP2_CTRL & (~bUEP_T_TOG) | UEP_T_RES_STALL;
                                              else
                                                 UEP2_CTRL = UEP2_CTRL & (~bUEP_R_TOG) | UEP_R_RES_STALL;
                                              break;     
                                         default:
                                              bRequestError=1;
                                      }
                                   }
                                   else bRequestError=1;
                               }
                               else bRequestError=1;
                               break;
                          case USB_GET_STATUS: //todo code here still missing
                               EP0_Buffer[0] = 0x00;
                               EP0_Buffer[1] = 0x00;
                               len = 2;
                               break;
                          default:
                               bRequestError=1; 
                               //break;
                       }
                    }
                    else  bRequestError=1;
                 }
                 else bRequestError=1;
        
                 if(bRequestError)
                 {
                    UEP0_CTRL = bUEP_R_TOG | bUEP_T_TOG | UEP_R_RES_STALL | UEP_T_RES_STALL;//STALL
                    break;
                 }
                 else if(len <= EP0_SIZE) UEP0_T_LEN = len;
                 else                     UEP0_T_LEN = 0;
                 UEP0_CTRL = bUEP_R_TOG | bUEP_T_TOG | UEP_R_RES_ACK | UEP_T_RES_ACK;
              }
              break;
         case UIS_TOKEN_IN | 0:
              {  //status out or datastagein todo check dir 
                 switch(Setup.bRequest)
                 {
                    case USB_GET_DESCRIPTOR:
                         len = Setup.wLength >= EP0_SIZE ? EP0_SIZE
                                                         : Setup.wLength;
                         //memcpy( EP0_Buffer, pDescr, len );
                         //pDescr += len;  
                         for (i=0; i<len; i++) {EP0_Buffer[i] = *pDescr++;}
                         Setup.wLength -= len;
                         UEP0_T_LEN     = len;
                         UEP0_CTRL     ^= bUEP_T_TOG; 
                         break;
                    case USB_SET_ADDRESS:
                         USB_DEV_AD = (USB_DEV_AD & bUDA_GP_BIT) | (Setup.wValueLo & 0x7F);
                         UEP0_CTRL  = UEP_R_RES_ACK | UEP_T_RES_NAK;
                         UsbState   = USB_STATE_ADDRESSED;
                         break;
                    default:
                         UEP0_T_LEN = 0;   
                         UEP0_CTRL = UEP_R_RES_ACK | UEP_T_RES_NAK;
                         //break;
                 }
              }
              break;
         case UIS_TOKEN_OUT | 0:
              {  // status in or datastage out todo check dir
                 UEP0_T_LEN = 0;  
                 UEP0_CTRL |= UEP_R_RES_ACK | UEP_T_RES_NAK; 
              }
//            break;
//       default:
//            break;
      }
      UIF_TRANSFER = 0;
   }   
   if(UIF_BUS_RST)
   {  // usbreset
      UsbState     = USB_STATE_RESET;  
      UEP0_CTRL    = UEP_R_RES_ACK | UEP_T_RES_NAK;
      //todo disable ep1..ep4
      USB_DEV_AD   = 0x00;
      UIF_SUSPEND  = 0;
      UIF_TRANSFER = 0;
      UIF_BUS_RST  = 0;
      //todo reset the bufctls
      UsbConfig     = 0;
      bUsb0InReady  = 1;
 
   }

   if (UIF_SUSPEND)
   {  // suspend
       UIF_SUSPEND = 0;
       if ( USB_MIS_ST & bUMS_SUSPEND )
       {
           while ( XBUS_AUX & bUART0_TX )
           {
              ;
           }
           SAFE_MOD  = 0x55;
           SAFE_MOD  = 0xAA;
           WAKE_CTRL = bWAK_BY_USB | bWAK_RXD0_LO | bWAK_RXD1_LO;
           UsbState |= USB_STATE_SUSPEND; //a bit stupid anyway
           PCON     |= PD;
           UsbState &=  ~USB_STATE_SUSPEND;
           SAFE_MOD  = 0x55;
           SAFE_MOD  = 0xAA;
           WAKE_CTRL = 0x00;
       }
   }
   else
   {  // ignore all others
      USB_INT_FG = 0xFF;
   }
}


extern uint8_t ReadDataFlash(uint8_t addr, uint8_t *value);



void main(void)
{

   if (ReadDataFlash(0x00,&CfgSense))   CfgSense   = 0; //sense option
   if (ReadDataFlash(0x01,&CfgOutStat)) CfgOutStat = 0; //running statua option  
   if (ReadDataFlash(0x02,&CfgUart1))   CfgUart1   = 0; //uart option
   
   if (CHIP_ID==ID_CH551) CfgUart1=0; // override it on CH551
   InitDevice();
   
   EA = 1;
   while(1)
   {
      if(UsbConfig)  // wait until configured
      {
         TransferMidi2Usb0();
         
         if (bActiveSense)
         {  // at this point the rx fifo is empty
         	  // so the transmission must be started manually
         	  bActiveSense=0;
            if(CfgSense) // just do something if activated;
            {
               EA=0;
        	     MidiTxBuffer[MidiTxCtl.rd++] = 0xFE;
               MidiTxCtl.rd&=MIDIBUF_MASK;
               MidiTxCtl.level++;	     
               bTx0Empty= 0;
               U1TI=1;
               EA=1; 
            } 
         }	
      }
   }
}
